import { GraphType, GraphTypeD3, LinkType, LinkTypeD3, NodeType, NodeTypeD3 } from '../types';
import { dataColors, visualizationColors } from 'config';
import { useEffect, useImperativeHandle, useMemo, useRef, useState } from 'react';
import {
  Application,
  Color,
  Container,
  FederatedPointerEvent,
  Graphics,
  IPointData,
  Sprite,
  Text,
  Texture,
  Resource,
  RenderTexture,
} from 'pixi.js';
import { useAppDispatch, useML, useSearchResultData } from '../../../../data-access';
import { NLPopUp } from './NLPopup';
import { hslStringToHex, nodeColor } from './utils';
import { CytoscapeLayout, GraphologyLayout, LayoutFactory, Layouts, GraphologyForceAtlas2Webworker } from '../../../../graph-layout';
import { MultiGraph } from 'graphology';
import { Viewport } from 'pixi-viewport';
import { NodelinkVisProps } from '../nodelinkvis';
import { Tooltip, TooltipContent, TooltipTrigger } from '@graphpolaris/shared/lib/components/tooltip';
import { MovedEvent } from 'pixi-viewport/dist/types';
import { VisualizationTooltip } from '@graphpolaris/shared/lib/components/VisualizationTooltip';
import { nodeColorHex } from './utils';
import { Theme } from '@graphpolaris/shared/lib/data-access/store/configSlice';
import { useConfig } from '@graphpolaris/shared/lib/data-access/store';

type Props = {
  onClick: (event?: { node: NodeTypeD3; pos: IPointData }) => void;
  // onHover: (data: { node: NodeType; pos: IPointData }) => void;
  // onUnHover: (data: { node: NodeType; pos: IPointData }) => void;
  highlightNodes: NodeType[];
  configuration: NodelinkVisProps;
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  graph: GraphType;
  layoutAlgorithm: string;
  showPopupsOnHover: boolean;
};

type LayoutState = 'reset' | 'running' | 'paused';

//////////////////
// MAIN COMPONENT
//////////////////

export const NLPixi = (props: Props) => {
  const [quickPopup, setQuickPopup] = useState<{ node: NodeType; pos: IPointData } | undefined>();
  const [popups, setPopups] = useState<{ node: NodeTypeD3; pos: IPointData }[]>([]);

  const globalConfig = useConfig();

  useEffect(() => {
    update();
  }, [globalConfig.theme]);

  const app = useMemo(
    () =>
      new Application({
        backgroundAlpha: 0,
        antialias: true,
        autoDensity: true,
        eventMode: 'auto',
        resolution: window.devicePixelRatio || 2,
      }),
    [],
  );
  const nodeLayer = useMemo(() => new Container(), []);
  const linkLabelLayer = useMemo(() => {
    const container = new Container();
    container.alpha = 0;
    container.renderable = false;
    return container;
  }, []);
  const nodeLabelLayer = useMemo(() => {
    const container = new Container();
    container.alpha = 0;
    container.renderable = false;
    return container;
  }, []);

  const nodeMap = useRef(new Map<string, Sprite>());
  const linkGfx = new Graphics();
  const linkLabelMap = useRef(new Map<string, Text>());
  const nodeLabelMap = useRef(new Map<string, Text>());
  const viewport = useRef<Viewport>();
  const layoutState = useRef<LayoutState>('reset');
  const layoutStoppedCount = useRef(0);
  const ref = useRef<HTMLDivElement>(null);
  const mouseInCanvas = useRef<boolean>(false);
  const [dragging, setDragging] = useState<boolean>(false);
  const isSetup = useRef(false);
  const ml = useML();
  const searchResults = useSearchResultData();
  const graph = useRef<GraphTypeD3>({ nodes: [], links: [] });

  const layoutAlgorithm = useRef<CytoscapeLayout | GraphologyLayout>(new LayoutFactory().createLayout(Layouts.DAGRE));

  // const cull = new Cull();
  // let cullDirty = useRef(true);

  const _glyphTexture = RenderTexture.create();
  const _selectedTexture = RenderTexture.create();

  const getTexture = (renderTexture: RenderTexture, selected: boolean = false): RenderTexture => {
    let size = config.NODE_RADIUS * (1 / responsiveScale);
    let lineWidth = selected ? 12 : 6;
    renderTexture.resize(size + lineWidth, size + lineWidth);

    const graphics = new Graphics();
    graphics.lineStyle(lineWidth, 0x4e586a);
    graphics.beginFill(0xffffff, 1);

    if (props.configuration.shapes?.shape == 'circle') {
      graphics.drawCircle(size / 2 + lineWidth / 2, size / 2 + lineWidth / 2, size / 2);
    } else {
      graphics.drawRect(lineWidth, lineWidth, size - lineWidth, size - lineWidth);
    }
    graphics.endFill();

    app.renderer.render(graphics, { renderTexture });

    return renderTexture;
  };

  // Pixi viewport zoom scale, but discretized to single decimal.
  const [responsiveScale, setResponsiveScale] = useState(1);

  useEffect(() => {
    if (nodeMap.current.size === 0) return;

    graph.current.nodes.forEach((node) => {
      const sprite = nodeMap.current.get(node._id) as Sprite;
      const nodeMeta = props.graph.nodes[node._id];
      sprite.texture = (sprite as any).selected ? selectedTexture : glyphTexture;

      // To calculate the scale, we:
      // 1) Determine the node radius, with a minimum of 5. If not available, we default to 5.
      // 2) Get the ratio with respect to the typical size of the node (divide by NODE_RADIUS).
      // 3) Scale this ratio by the current scale factor.
      let scale = (Math.max(nodeMeta.radius || 5, 5) / config.NODE_RADIUS) * 2;
      scale *= responsiveScale;
      sprite.scale.set(scale, scale);
    });

    if (graph.current.nodes.length > config.LABEL_MAX_NODES) return;

    // Change font size at specific scale intervals
    const fontSize =
      responsiveScale <= 0.1 ? 15 : responsiveScale <= 0.2 ? 22.5 : responsiveScale <= 0.4 ? 30 : responsiveScale <= 0.6 ? 37.5 : 45;

    const strokeWidth = fontSize / 2;
    linkLabelMap.current.forEach((text) => {
      text.style.fontSize = fontSize;
      text.style.strokeThickness = strokeWidth;
      text.resolution = Math.ceil(0.5 / responsiveScale);
    });

    nodeLabelMap.current.forEach((text) => {
      text.style.fontSize = fontSize * (2 / 3);
      text.resolution = Math.ceil(1 / responsiveScale);
    });

    graph.current.nodes.forEach((node: any) => {
      updateNodeLabel(node);
    });
  }, [responsiveScale, props.configuration.shapes?.shape]);

  const [config, setConfig] = useState({
    width: 1000,
    height: 1000,

    LABEL_MAX_NODES: 1000,

    LAYOUT_ALGORITHM: Layouts.FORCEATLAS2WEBWORKER,

    NODE_RADIUS: 70,
    // NODE_BORDER_LINE_WIDTH: 1.0,
    // NODE_BORDER_LINE_WIDTH_SELECTED: 5.0, // if selected and normal width are different the thicker line will be still in the gfx
    // NODE_BORDER_COLOR_DEFAULT: dataColors.neutral[70],
    // NODE_BORDER_COLOR_SELECTED: dataColors.orange[60],

    LINE_COLOR_DEFAULT: dataColors.neutral[40],
    LINE_COLOR_SELECTED: visualizationColors.GPSelected.colors[1],
    LINE_COLOR_ML: dataColors.blue[60],
    LINE_WIDTH_DEFAULT: 0.8,
  });

  const glyphTexture = useMemo(() => {
    return getTexture(_glyphTexture);
  }, [responsiveScale, props.configuration.shapes?.shape]);

  const selectedTexture = useMemo(() => {
    return getTexture(_selectedTexture, true);
  }, [responsiveScale, props.configuration.shapes?.shape]);

  useEffect(() => {
    setConfig((lastConfig) => {
      return {
        ...lastConfig,
        LAYOUT_ALGORITHM: (props.layoutAlgorithm as Layouts) || lastConfig.LAYOUT_ALGORITHM,
      };
    });
  }, [props.layoutAlgorithm, props.configuration]);

  const imperative = useRef<any>(null);

  const mouseClickThreshold = 200; // Time between mouse up and down events that is considered a click, and not a drag.

  useImperativeHandle(imperative, () => ({
    onMouseDown(event: FederatedPointerEvent) {
      if (props.configuration.showPopUpOnHover) return;

      (event as any).mouseDownTimeStamp = event.timeStamp;
      setDragging(true);
    },

    onMouseUpNode(event: FederatedPointerEvent) {
      if (props.configuration.showPopUpOnHover) return;

      // If its a short click (not a drag) on the stage but not on a node: clear the selection and remove all popups.
      const holdDownTime = event.timeStamp - (event as any).mouseDownTimeStamp;
      if (holdDownTime > mouseClickThreshold) {
        return;
      }

      const sprite = event.target as Sprite;
      const node = (sprite as any).node as NodeTypeD3;

      if (event.shiftKey) {
        setPopups([...popups, { node: node, pos: toGlobal(node) }]);
      } else {
        setPopups([{ node: node, pos: toGlobal(node) }]);
        for (const popup of popups) {
          const sprite = nodeMap.current.get(popup.node._id) as Sprite;
          sprite.texture = glyphTexture;
          (sprite as any).selected = false;
        }
      }

      sprite.texture = selectedTexture;
      (sprite as any).selected = true;
      setDragging(false);

      props.onClick({ node: node, pos: toGlobal(node) });

      event.stopPropagation();
    },

    onMouseUpStage(event: FederatedPointerEvent) {
      if (props.configuration.showPopUpOnHover) return;

      // If its a short click (not a drag) on the stage but not on a node: clear the selection and remove all popups.
      const holdDownTime = event.timeStamp - (event as any).mouseDownTimeStamp;
      if (holdDownTime < mouseClickThreshold) {
        for (const popup of popups) {
          const sprite = nodeMap.current.get(popup.node._id) as Sprite;
          sprite.texture = glyphTexture;
          (sprite as any).selected = false;
        }
        setPopups([]);
        props.onClick();
      }
    },

    onHover(event: FederatedPointerEvent) {
      if (!props.configuration.showPopUpOnHover) return;

      const sprite = event.target as Sprite;
      const node = (sprite as any).node as NodeTypeD3;
      if (
        mouseInCanvas.current &&
        viewport?.current &&
        !viewport?.current?.pause &&
        node &&
        popups.filter((p) => p.node._id === node._id).length === 0
      ) {
        setQuickPopup({ node: props.graph.nodes[node._id], pos: toGlobal(node) });
      }
    },
    onUnHover() {
      if (!props.configuration.showPopUpOnHover) return;

      setQuickPopup(undefined);
    },
    onMoved(event: MovedEvent) {
      if (props.configuration.showPopUpOnHover) return;

      for (const popup of popups) {
        if (popup.node.x == null || popup.node.y == null) continue;
        popup.pos.x = event.viewport.transform.position.x + popup.node.x * event.viewport.scale.x;
        popup.pos.y = event.viewport.transform.position.y + popup.node.y * event.viewport.scale.y;
      }
      setPopups([...popups]);
    },
    onZoom(event: FederatedPointerEvent) {
      const scale = viewport.current!.transform.scale.x;

      if (scale > 2) {
        const scale = 1 / viewport.current!.scale.x; // starts from 0.5 down to 0.
        setResponsiveScale(scale < 0.05 ? 0.1 : scale < 0.1 ? 0.2 : scale < 0.2 ? 0.4 : scale < 0.3 ? 0.6 : 0.8);
      } else {
        setResponsiveScale(1);
      }

      if (graph.current.nodes.length < config.LABEL_MAX_NODES) {
        linkLabelLayer.alpha = scale > 2 ? Math.min(1, (scale - 2) * 3) : 0;

        if (linkLabelLayer.alpha > 0) {
          linkLabelLayer.renderable = true;
        } else {
          linkLabelLayer.renderable = false;
        }

        nodeLabelLayer.alpha = scale > 5 ? Math.min(1, (scale - 5) * 3) : 0;
        if (nodeLabelLayer.alpha > 0) {
          nodeLabelLayer.renderable = true;
        } else {
          nodeLabelLayer.renderable = false;
        }
      }
    },

    getLinkWidth() {
      return props.configuration.edges.width.width || config.LINE_WIDTH_DEFAULT;
    },

    getBackgroundColor() {
      // Colors corresponding to .bg-light class
      return globalConfig.theme === Theme.dark ? 0x121621 : 0xffffff;
    },
  }));

  function resize() {
    const width = ref?.current?.clientWidth || 1000;
    const height = ref?.current?.clientHeight || 1000;
    app.renderer.resize(width, height);
    if (viewport.current) {
      viewport.current.screenWidth = width;
      viewport.current.worldWidth = width;
      viewport.current.worldHeight = height;
      viewport.current.screenHeight = height;
    }
    app.render();
  }
  useEffect(() => {
    if (!ref.current) return;
    const resizeObserver = new ResizeObserver(() => {
      resize();
    });
    resizeObserver.observe(ref.current);
    return () => resizeObserver.disconnect(); // clean up
  }, []);

  useEffect(() => {
    if (ref.current && ref.current.children.length === 0) {
      ref.current.appendChild(app.view as HTMLCanvasElement);
      resize();
    }
  }, [ref]);

  function toGlobal(node: NodeTypeD3): IPointData {
    if (viewport?.current) {
      // const rect = ref.current?.getBoundingClientRect();
      const rect = { x: 0, y: 0 };
      const x = (rect?.x || 0) + (node.x || 0);
      const y = (rect?.y || 0) + (node.y || 0);
      return viewport.current.toScreen(x, y);
    } else return { x: 0, y: 0 };
  }

  const updateNode = (node: NodeTypeD3) => {
    const gfx = nodeMap.current.get(node._id);
    if (!gfx) return;

    // Update texture when selected
    const nodeMeta = props.graph.nodes[node._id];
    const texture = (gfx as any).selected ? selectedTexture : glyphTexture;
    gfx.texture = texture;

    // Cluster colors
    if (nodeMeta?.cluster) {
      gfx.tint = nodeMeta.cluster >= 0 ? nodeColor(nodeMeta.cluster) : 0x000000;
    } else {
      gfx.tint = nodeColor(nodeMeta.type);
    }

    gfx.position.set(node.x, node.y);

    // if (!item.position) {
    //   item.position = new Point(node.x, node.y);
    // } else {
    //   item.position.set(node.x, node.y);
    // }
    // Update attributes position if they exist
    // if (node.gfxAttributes) {
    //   const x = node.x - node.gfxAttributes.width / 2;
    //   const y = node.y - node.gfxAttributes.height - 20;
    //   if (!node.gfxAttributes?.position) node.gfxAttributes.position = new Point(x, y);
    //   else {
    //     node.gfxAttributes.position.set(x, y);
    //   }
    // }
  };

  const getNodeLabel = (nodeMeta: NodeType) => {
    return nodeMeta.label;
  };

  const createNode = (node: NodeTypeD3, selected?: boolean) => {
    const nodeMeta = props.graph.nodes[node._id];

    // check if node is already drawn, and if so, delete it
    if (node && node?._id && nodeMap.current.has(node._id)) {
      nodeMap.current.delete(node._id);
    }
    // Do not draw node if it has no position
    if (node.x === undefined || node.y === undefined) return;

    let sprite: Sprite;
    const texture = glyphTexture;
    sprite = new Sprite(texture);

    sprite.tint = nodeColor(nodeMeta.type);
    const scale = (Math.max(nodeMeta.radius || 5, 5) / config.NODE_RADIUS) * 2;
    sprite.scale.set(scale, scale);
    sprite.anchor.set(0.5, 0.5);
    sprite.cullable = true;

    sprite.eventMode = 'static';
    sprite.on('mousedown', (e) => imperative.current.onMouseDown(e));
    sprite.on('mouseup', (e) => imperative.current.onMouseUpNode(e));
    sprite.on('mouseover', (e) => imperative.current.onHover(e));
    sprite.on('mouseout', (e) => imperative.current.onUnHover(e));

    nodeMap.current.set(node._id, sprite);
    nodeLayer.addChild(sprite);

    updateNode(node);
    (sprite as any).node = node;

    // Node label
    const attribute = getNodeLabel(nodeMeta);
    const text = new Text(attribute, {
      fontSize: 20,
      fill: 0xffffff,
      wordWrap: true,
      wordWrapWidth: 65,
      align: 'center',
    });
    text.eventMode = 'none';
    text.cullable = true;
    text.anchor.set(0.5, 0.5);
    text.scale.set(0.1, 0.1);
    nodeLabelMap.current.set(node._id, text);
    nodeLabelLayer.addChild(text);

    updateNodeLabel(node);

    return sprite;
  };

  const createLinkLabel = (link: LinkTypeD3) => {
    // check if link is already drawn, and if so, delete it
    if (link && link?._id && linkLabelMap.current.has(link._id)) {
      linkLabelMap.current.delete(link._id);
    }

    const linkMeta = props.graph.links[link._id];

    const text = new Text(linkMeta.name, {
      fontSize: 60,
      fill: config.LINE_COLOR_DEFAULT,
      stroke: imperative.current.getBackgroundColor(),
      strokeThickness: 30,
    });
    text.cullable = true;
    text.anchor.set(0.5, 0.5);
    text.scale.set(0.1, 0.1);
    linkLabelMap.current.set(link._id, text);
    linkLabelLayer.addChild(text);

    updateLinkLabel(link);

    return text;
  };

  const updateLink = (link: LinkTypeD3) => {
    if (!props.graph || nodeMap.current.size === 0) return;
    const linkMeta = props.graph.links[link._id];

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      return;
    }

    let sourceId = '';
    let targetId = '';
    let source: NodeTypeD3 | undefined;
    let target: NodeTypeD3 | undefined;

    if (typeof _source === 'string') {
      sourceId = link.source as string;
      targetId = link.target as string;
      source = nodeMap.current.get(sourceId) as NodeTypeD3 | undefined;
      target = nodeMap.current.get(targetId) as NodeTypeD3 | undefined;
    } else {
      source = link.source as NodeTypeD3;
      target = link.target as NodeTypeD3;
      sourceId = source._id;
      targetId = target._id;
    }
    if (!source || !target) {
      console.error('source or target not found', source, target, sourceId, targetId);
      return;
    }

    // let color = link.color || 0x000000;
    let color = config.LINE_COLOR_DEFAULT;
    let style = imperative.current.getLinkWidth();
    let alpha = linkMeta.alpha || 1;
    if (linkMeta.mlEdge) {
      color = config.LINE_COLOR_ML;
      if (linkMeta.value > ml.communityDetection.jaccard_threshold) {
        style = linkMeta.value * 1.8;
      } else {
        style = 0;
        alpha = 0.2;
      }
    } else if (props.highlightedLinks && props.highlightedLinks.includes(linkMeta)) {
      if (linkMeta.mlEdge && ml.communityDetection.jaccard_threshold) {
        if (linkMeta.value > ml.communityDetection.jaccard_threshold) {
          color = dataColors.magenta[50];
          // 0xaa00ff;
          style = linkMeta.value * 1.8;
        }
      } else {
        color = dataColors.red[70];
        // color = 0xff0000;
        style = 1.0;
      }
    } else if (props.currentShortestPathEdges && props.currentShortestPathEdges.includes(linkMeta)) {
      color = dataColors.green[50];
      // color = 0x00ff00;
      style = 3.0;
    }

    // Conditional alpha for search results
    if (searchResults.nodes.length > 0 || searchResults.edges.length > 0) {
      // FIXME: searchResults.edges should be a hashmap to improve performance.
      const isLinkInSearchResults = searchResults.edges.some((resultEdge) => resultEdge.id === link._id);
      alpha = isLinkInSearchResults ? 1 : 0.05;
    }

    linkGfx
      .lineStyle(style, hslStringToHex(color), alpha)
      .moveTo(source.x || 0, source.y || 0)
      .lineTo(target.x || 0, target.y || 0);
  };

  const updateLinkLabel = (link: LinkTypeD3) => {
    if (graph.current.nodes.length > config.LABEL_MAX_NODES || viewport.current!.transform.scale.x < 2) return;

    const text = linkLabelMap.current.get(link._id);
    if (!text) return;

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      return;
    }

    const source = nodeMap.current.get(link.source as string) as Sprite;
    const target = nodeMap.current.get(link.target as string) as Sprite;

    text.x = (source.x + target.x) / 2;
    text.y = (source.y + target.y) / 2;

    const length = Math.hypot(target.x - source.x, target.y - source.y);

    // Skip rendering labels on very short edges
    if (length < text.width + 10) {
      // 10 to account for size of node
      text.alpha = 0;
      return;
    } else {
      text.alpha = 1;
    }

    const rads = Math.atan2(target.y - source.y, target.x - source.x);
    text.rotation = rads;

    const degrees = Math.abs(text.angle % 360);

    // Rotate edge labels to always be legible
    if (degrees > 90 && degrees < 270) {
      text.rotation = rads + Math.PI;
    } else {
      text.rotation = rads;
    }

    text.style.stroke = imperative.current.getBackgroundColor();
  };

  const updateNodeLabel = (node: NodeTypeD3) => {
    if (graph.current.nodes.length > config.LABEL_MAX_NODES || viewport.current!.transform.scale.x < 2) return;
    const text = nodeLabelMap.current.get(node._id) as Text | undefined;
    if (text == null) return;

    if (node.x) text.x = node.x;
    if (node.y) text.y = node.y;

    const nodeMeta = props.graph.nodes[node._id];
    const originalText = getNodeLabel(nodeMeta);

    text.text = originalText; // This is required to ensure the text size check (next line) works

    if (text.width / text.scale.x <= 90 && text.height / text.scale.y <= 90) {
      text.text = originalText;
    } else {
      // Change character limit at specific scale intervals
      const charLimit = responsiveScale > 0.2 ? 15 : responsiveScale > 0.1 ? 30 : 75;
      text.text = `${originalText.slice(0, charLimit)}…`;
    }

    text.alpha = text.width / text.scale.x <= 90 && text.height / text.scale.y <= 90 ? 1 : 0;
  };

  // const text = linkLabelMap.current.get(link._id);
  //   if (!text) return;

  //   const source = link.source as NodeTypeD3;
  //   const target = link.target as NodeTypeD3;

  //   if (source.x == null || source.y == null || target.x == null || target.y == null) return;

  //   text.x = (source.x + target.x) / 2;
  //   text.y = (source.y + target.y) / 2;

  //   const rads = Math.atan2(target.y - source.y, target.x - source.x);
  //   const degrees = Math.abs(text.angle % 360);

  //   // Rotate edge labels to always be legible
  //   if (degrees > 90 && degrees < 270) {
  //     text.rotation = rads + Math.PI;
  //   } else {
  //     text.rotation = rads;
  //   }

  useEffect(() => {
    return () => {
      nodeMap.current.clear();
      linkLabelMap.current.clear();
      linkGfx.clear();
      nodeLayer.removeChildren();
      linkLabelLayer.removeChildren();
      nodeLabelLayer.removeChildren();

      const layout = layoutAlgorithm.current as GraphologyForceAtlas2Webworker;
      if (layout?.cleanup != null) layout.cleanup();
    };
  }, []);

  useEffect(() => {
    if (props.graph && ref.current && ref.current.children.length > 0 && imperative.current) {
      if (isSetup.current === false) setup();
      else update(false);
    }
  }, [config]);

  useEffect(() => {
    if (props.graph) {
      graph.current.nodes.forEach((node) => {
        const gfx = nodeMap.current.get(node._id);
        if (!gfx) return;
        const isNodeInSearchResults = searchResults.nodes.some((resultNode) => resultNode.id === node._id);

        gfx.alpha = isNodeInSearchResults || searchResults.nodes.length === 0 ? 1 : 0.05;
      });
    }
  }, [searchResults]);

  const tick = (delta: number) => {
    if (layoutState.current === 'paused') return;
    if (layoutState.current === 'reset') layoutStoppedCount.current = 0;

    if (props.graph) {
      if (!layoutAlgorithm.current) return;

      let stopped = 0;

      const widthHalf = app.renderer.width / 2;
      const heightHalf = app.renderer.height / 2;
      graph.current.nodes.forEach((node, i) => {
        if (!layoutAlgorithm.current) return;
        const gfx = nodeMap.current.get(node._id);
        if (!gfx || node.x === undefined || node.y === undefined) return;

        const position = layoutAlgorithm.current.getNodePosition(node._id);

        if (!position || Math.abs(node.x - position.x - widthHalf) + Math.abs(node.y - position.y - heightHalf) < 1) {
          stopped += 1;
          return;
        }
        if (layoutAlgorithm.current.provider === 'Graphology') {
          // this is a dirty hack to fix the graphology layout being out of bounds
          node.x = position.x + widthHalf;
          node.y = position.y + heightHalf;
        } else {
          node.x = position.x;
          node.y = position.y;
        }

        gfx.position.copyFrom(node as IPointData);

        updateNodeLabel(node);
      });

      if (stopped === graph.current.nodes.length) {
        layoutStoppedCount.current = layoutStoppedCount.current + 1;
        if (layoutStoppedCount.current > 1000) {
          layoutState.current = 'paused';
          console.debug('NL layout paused');
        }
      } else {
        layoutStoppedCount.current = 0;
      }
      if (layoutState.current === 'reset') {
        layoutState.current = 'running';
      }

      // Draw the links
      linkGfx.clear();
      linkGfx.beginFill();
      graph.current.links.forEach((link: any) => {
        updateLink(link);
        updateLinkLabel(link);
      });
      linkGfx.endFill();
    }
  };

  const update = (forceClear = false) => {
    if (!props.graph || !ref.current) return;

    if (props.graph) {
      if (forceClear) {
        nodeMap.current.clear();
        linkGfx.clear();
        nodeLayer.removeChildren();
        linkLabelLayer.removeChildren();
        nodeLabelLayer.removeChildren();
      }

      nodeMap.current.forEach((gfx, id) => {
        if (!graph.current.nodes.find((node) => node._id === id)) {
          nodeLayer.removeChild(gfx);
          gfx.destroy();
          nodeMap.current.delete(id);
        }
      });

      linkLabelMap.current.forEach((text, id) => {
        if (!graph.current.links.find((link) => link._id === id)) {
          linkLabelLayer.removeChild(text);
          text.destroy();
          linkLabelMap.current.delete(id);
        }
      });

      linkGfx.clear();

      graph.current.nodes.forEach((node) => {
        if (!forceClear && nodeMap.current.has(node._id)) {
          const old = nodeMap.current.get(node._id);

          node.x = old?.x || node.x;
          node.y = old?.y || node.y;
          updateNode(node);
          updateNodeLabel(node);
        } else {
          createNode(node);
        }
      });

      if (graph.current.nodes.length < config.LABEL_MAX_NODES) {
        graph.current.links.forEach((link) => {
          if (!forceClear && linkLabelMap.current.has(link._id)) {
            updateLinkLabel(link);
          } else {
            createLinkLabel(link);
          }
        });
      }

      // // update text colour (written after nodes so that text appears on top of nodes)
      //   nodes.forEach((node: NodeType) => {
      //   if (node.gfxAttributes !== undefined) {
      //       const selected = node.selected === true;
      //       node.gfxAttributes.destroy();
      //       createAttributes(node);
      //       if (selected) {
      //       showAttributes(node);
      //       }
      //   }
      //   });
      // console.log(nodes.children);

      // force.startSimulation(props.graph, ref.current.getBoundingClientRect());
      // force.simulation.on('tick', () => {});
      app.ticker.add(tick);
      layoutState.current = 'reset';
      if (forceClear) setupLayout(forceClear);
    }
  };

  /**
   * SetNodeGraphics is an initializing function. It attaches the nodes and links to the simulation.
   * It creates graphic objects and adds these to the PIXI containers. It also clears both of these of previous nodes and links.
   * @param graph The graph returned from the database and that is parsed into a nodelist and edgelist.
   */
  const setup = async () => {
    nodeLayer.removeChildren();
    linkLabelLayer.removeChildren();
    app.stage.removeChildren();

    if (!props.graph) throw Error('Graph is undefined');

    //Setup d3 graph structure
    graph.current = {
      nodes: Object.values(props.graph.nodes).map((n) => ({ _id: n._id, x: n.defaultX, y: n.defaultY })),
      links: Object.values(props.graph.links).map((l) => ({
        _id: l.id,
        source: l.source,
        target: l.target,
      })),
    };

    const size = ref.current?.getBoundingClientRect();
    viewport.current = new Viewport({
      screenWidth: size?.width || 1000,
      screenHeight: size?.height || 1000,
      worldWidth: size?.width || 1000,
      worldHeight: size?.height || 1000,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });
    app.stage.addChild(viewport.current);
    // activate plugins
    viewport.current.drag().pinch().wheel({ smooth: 2 }).animate({}).decelerate({ friction: 0.75 });

    viewport.current.addChild(linkGfx);
    viewport.current.addChild(linkLabelLayer);
    viewport.current.addChild(nodeLayer);
    viewport.current.addChild(nodeLabelLayer);
    viewport.current.on('moved', (event) => {
      imperative.current.onMoved(event);
    });
    viewport.current.on('drag-end', (event) => {
      setDragging(false);
    });
    viewport.current.on('zoomed', (event) => {
      imperative.current.onZoom(event);
    });

    app.stage.eventMode = 'dynamic';
    app.stage.on('mousedown', (e) => imperative.current.onMouseDown(e));
    app.stage.on('mouseup', (e) => imperative.current.onMouseUpStage(e));

    nodeMap.current.clear();
    linkGfx.clear();
    update(true);
    isSetup.current = true;
  };

  const setupLayout = (forceClear: boolean) => {
    const layoutFactory = new LayoutFactory();
    layoutAlgorithm.current = layoutFactory.createLayout(config.LAYOUT_ALGORITHM);

    if (!layoutAlgorithm) throw Error('LayoutAlgorithm is undefined');

    const graphologyGraph = new MultiGraph();
    graph.current.nodes.forEach((node) => {
      if (forceClear) graphologyGraph.addNode(node._id, { size: props.graph.nodes[node._id].radius || 5 });
      else graphologyGraph.addNode(node._id, { size: props.graph.nodes[node._id].radius || 5, x: node.x || 0, y: node.y || 0 });
    });

    graph.current.links.forEach((link) => {
      graphologyGraph.addEdge(link.source, link.target);
    });
    const boundingBox = { x1: 0, x2: app.renderer.screen.width, y1: 0, y2: app.renderer.screen.height };

    if (forceClear) {
      const startingLayout = layoutFactory.createLayout('Graphology_random');
      startingLayout.layout(graphologyGraph, boundingBox);
    }
    layoutAlgorithm.current.layout(graphologyGraph, boundingBox);
  };

  return (
    <>
      {popups.map((popup) => (
        <Tooltip key={popup.node._id} open={true} interactive={!dragging} boundaryElement={ref} showArrow={true}>
          <TooltipTrigger x={popup.pos.x} y={popup.pos.y} />
          <TooltipContent>
            <VisualizationTooltip name={popup.node._id} colorHeader={nodeColorHex(props.graph.nodes[popup.node._id].type)}>
              <NLPopUp data={props.graph.nodes[popup.node._id].attributes} />
            </VisualizationTooltip>
          </TooltipContent>
        </Tooltip>
      ))}
      {quickPopup != null && (
        <Tooltip key={quickPopup.node._id} open={true} boundaryElement={ref} showArrow={true}>
          <TooltipTrigger x={quickPopup.pos.x} y={quickPopup.pos.y} />
          <TooltipContent>
            <VisualizationTooltip name={quickPopup.node._id} colorHeader={nodeColorHex(props.graph.nodes[quickPopup.node._id].type)}>
              <NLPopUp data={props.graph.nodes[quickPopup.node._id].attributes} />
            </VisualizationTooltip>
          </TooltipContent>
        </Tooltip>
      )}
      <div
        className="h-full w-full overflow-hidden"
        ref={ref}
        onMouseEnter={(e) => {
          mouseInCanvas.current = true;
        }}
        onMouseOut={(e) => {
          mouseInCanvas.current = false;
        }}
      ></div>
    </>
  );
};
